module stack
#(
    parameter N = 8, // number of bits
              S = 10, // stack size
              W = $clog2(S+1)  // number of addr bits
) (
    input          clk, rstb,
    input          push, pop,
    input  [N-1:0] w_data,
    output         empty, full,
    output [N-1:0] r_data
);

// signal declaration
reg  [W-1:0] top_reg, top_next, top_dec, top_inc;
reg          full_reg, empty_reg, full_next, empty_next;
wire         ram_we;
wire [W-1:0] ram_addr;
wire [N-1:0] ram_din, ram_dout;

// RAM
sp_ram_sync_read
#(
    .N   (N       ),
    .W   (W       )
) mem (
    .clk (clk     ),
    .we  (ram_we  ),
    .addr(ram_addr),
    .din (ram_din ),
    .dout(ram_dout)
);

assign ram_we = push & ~full_reg; // write when RAM is not full
assign ram_addr = push ? top_inc : top_reg;
assign ram_din = w_data;
assign r_data = (push&pop) ? w_data : ram_dout;

// registers
always @(posedge clk or negedge rstb) begin
    if (~rstb) begin
        top_reg   <= 2**W-1;
        full_reg  <= 0;
        empty_reg <= 1;
    end
    else begin
        top_reg   <= top_next;
        full_reg  <= full_next;
        empty_reg <= empty_next;
    end
end

// next-state logic
always @* begin
    top_dec = top_reg-1;
    top_inc = top_reg+1;
    // default values
    top_next = top_reg;
    full_next = full_reg;
    empty_next = empty_reg;
    case ({push, pop})
        // 2'b00: no operation
        2'b01: // pop (read)
            if (~empty_reg) begin
                top_next = top_dec;
                full_next = 0;
                if (top_dec==2**W-1) empty_next = 1;
            end
        2'b10: // push (write)
            if (~full_reg) begin
                top_next = top_inc;
                empty_next = 0;
                if (top_inc==S-1) full_next = 1;
            end
        // 2'b11: // push and pop (do nothing)
    endcase
end

assign full = full_reg;
assign empty = empty_reg;

endmodule